module Codeclimate
  module Collectors
    module Messages
      class Message
        include ActiveModel::Model
        include ActiveModel::Validations::HelperMethods
        include Collectors::Validations

        def self.attribute_metadata
          @attribute_metadata ||= {}
        end

        # Declare an attribute on a message type. This will declare an
        # +attr_accessor+, a validation that the attribute is of the specified
        # +type+, and include it in +#attributes+. If +optional+ is false a
        # validation for +presence+ is also added.
        def self.attribute(name, type, optional: false)
          name = name.to_sym

          attribute_metadata[name] = { type: type, optional: optional }

          attr_accessor name

          validates name, presence: !optional, type: { type: type }
        end

        # Calculate the type string to use in the JSON representation to
        # identify this message type.
        def self.json_type
          to_s.gsub(/^.+::(\w+)$/, "\\1")
        end

        # Messages can have ordering dependencies, e.g. a collector could emit
        # an Incident message, followed by several IncidentEvent messages.
        # IncidentEvent messages identify their associated incident via their
        # +incident_external_id+ attribute.
        #
        # Our backend distributes messages amongst parallel workers for
        # ingestion. To ensure messages with an ordering dependency are
        # processed in order by the same worker, we can return an appropriate
        # +ordering_keys+ array for the classes: e.g. Incident returns
        # +[:external_id]+ and IncidentEvent returns +[incident_external_id]+.
        #
        # Collectors are also responsible for emitting messages in the
        # appropriate order.
        def self.ordering_keys
          []
        end

        def ==(other)
          other.class == self.class && other.attributes == attributes
        end

        validate :validate_not_base_class

        # Return all the attributes of a message as a +Hash+
        def attributes
          Hash[
            self.class.attribute_metadata.keys.each.map do |name|
              [name, public_send(name)]
            end
          ]
        end

        # A +Hash+ for serializing the message as JSON.
        # Includes a +type+ key and an +attributes+ key.
        def as_json(*_opts)
          {
            type: self.class.json_type,
            attributes: json_attributes,
          }
        end

        protected

        def json_attributes
          Hash[attributes.map { |name, val| [name, val.as_json] }]
        end

        private

        def validate_not_base_class
          if instance_of?(Messages::Message)
            errors.add(:base, "base Message class is abstract")
          end
        end
      end
    end
  end
end

